import logging

from quiq.evaluate import evaluate, environment

class TestCase:
    """A Test case - it acts as the interface to the user and constructs and
    runs tesp code"""
    def __init__(self, *args):
        # Optional: The first argument can be a description, everything else
        # is an unevalauted texp.
        self.desc = None

        if isinstance(args[0], str):
            self.desc = args[0]
            args = args[1:]
        self.tests = args

    def __call__(self, func):
        self.func = func
        if not self.desc: self.desc = func.__name__

        # Bind the given function to 'F'
        env = environment.copy()
        env['F'] = self.func

        # Tesp code to run tests and return test results that look like:
        # (didPass Test actual)
        tesp = ['list']
        for test in self.tests:
            op, F, args, expected = test.operator, test.function, test.args, test.expected
            test = [op, [F, *args], expected]
            actual  = [F, *args]
            # Append a test result
            tesp += [['list', test, ['quote', test], actual]]

        results = evaluate(tesp, env)

        self._report(results)

        return func

    def _report(self, results):
        logging.basicConfig(format='%(message)s')
        log = logging.getLogger(__name__)

        log.warning(f"Test Case: {self.desc}")

        for didPass, (operator, (F, *args), expected), actual in results:
            line = "✅" if didPass else "❌"

            func_name = self.func.__name__
            args = ', '.join([str(a) for a in args])

            # Process lists specially:
            if isinstance(expected, list):
                expected = ', '.join([str(e) for e in expected[1:]]) # Remove the first element, 'list'
                expected = f"[{expected[:10]}...{expected[-10:]}]"

            line += f"  Original: {func_name}({args}) {operator} {expected}\n"
            if isinstance(actual, Exception):
                line += f"    Expanded: {repr(actual)}"
            else:
                line += f"    Expanded: {' ' * (len(func_name + args)) }{actual} {operator} {expected}"
            line += '\n'
            log.warning(line)



from quiq.test import ExceptionWrapper

# Types
Symbol = str
Number = (int, float)
Bool   = bool

# Maps functions to their implementation
environment = {
    # Basics
    'list' : lambda *args: list(args),
    'quote': 'quote',
    'car'  : lambda L    : L[0],
    'cdr'  : lambda L    : list(L[1:]),

    # Operators
    'and' : lambda *args: reduce(lambda x, y: x and y, args),
    'or'  : lambda *args: reduce(lambda x, y: x or  y, args),
    '=='  : lambda x, y : x == y,
    '!='  : lambda x, y : x != y,
    '<'   : lambda x, y : x <  y,
    '<='  : lambda x, y : x <= y,
    '>'   : lambda x, y : x >  y,
    '>='  : lambda x, y : x >= y,
    'isin': lambda x, y : x in y,
    'throws': lambda err, err_class: isinstance(err, err_class),
}


class UnknownSymbolError(Exception): ...


def evaluate(texp, env=environment):
    "Evaluate an expression in an environment."
    if isinstance(texp, Symbol):
        if texp not in env: raise UnknownSymbolError(f"Unknown symbol: `{texp}`")
        return env[texp]
    elif isinstance(texp, (Number, Bool)):
        return texp
    elif isinstance(texp, ExceptionWrapper):
        return texp.err_type

    # Functions
    else:
        func = evaluate(texp[0], env)
        if func == 'quote': return texp[1]
        args = [evaluate(t, env) for t in texp[1:]]

        # Try calling the function. Catch exception and return them for throw
        # to catch
        try:
            result = func(*args)
        except Exception as e:
            result = e
        return result



# A constant to represent the test being called
_F = 'F'

class ExceptionWrapper:
    """Wraps an Exception type so it can be type checked"""
    def __init__(self, err_type):
        self.err_type = err_type
    def __repr__(self):
        return repr(self.err_type)[8:-2]

class Test():
    """The basic unit for a test. It holds:
    - The function thats being tested
    - Arguments used to call the function
    - The operator to compare outputs
    - The expected value when the function is evaluated
    """
    def __init__(self, *args):
        self.function = _F          # The function to test
        self.args     = list(args)  # Args to evaluate the function with
        self.operator = None        # How we expect to interpret the value of the function
        self.expected = None        # What we expect to happen when the function is called

    def __repr__(self):
        res =  f"{self.function}("
        res += ', '.join([str(a) for a in self.args])
        res += ")"
        res += f" {self.operator}"
        res += f" {self.expected}"
        return res

    """ ----------------------------------------------------------------------
    OPERATORS
    -----------------------------------------------------------------------"""
    def __eq__(self, expected): self.operator, self.expected = '==', expected; return self
    def __ne__(self, expected): self.operator, self.expected = '!=', expected; return self
    def __lt__(self, expected): self.operator, self.expected = '<',  expected; return self
    def __le__(self, expected): self.operator, self.expected = '<=', expected; return self
    def __gt__(self, expected): self.operator, self.expected = '>',  expected; return self
    def __ge__(self, expected): self.operator, self.expected = '>=', expected; return self

    def isin(self, other):
        self.operator, self.expected = 'isin', ['list', *other]
        return self

    def throws(self, err=None):
        self.operator = 'throws'
        self.expected = ExceptionWrapper(err) if err else ExceptionWrapper(Exception)
        return self

# Alias
F = Test
